/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "packet.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "client/client.h"
#include "server/server.h"

packethandler_t PACKET_HANDLERS[] = {
  { NULL, NULL },
  { NULL, NULL },
  { packetDisconnectClientProcess, NULL },
  { packetPingClientProcess, packetPingServerProcess },
};

void packetInit(
  packet_t *packet,
  const packettype_t type,
  const uint32_t length
) {
  assertNotNull(packet, "Packet is NULL");

  memoryZero(packet, sizeof(packet_t));
  
  assertTrue(length > 0, "Packet length is 0");
  assertTrue(
    length <= sizeof(packetdata_t),
    "Packet length is too large"
  );

  packet->type = type;
  packet->length = length;
}

errorret_t packetClientProcess(
  const packet_t *packet,
  client_t *client
) {
  assertNotNull(packet, "Packet is NULL");
  assertNotNull(client, "Client is NULL");
  assertTrue(
    client->type == CLIENT_TYPE_NETWORKED,
    "Client is not networked"
  );
  assertIsMainThread("Client process must be on main thread");

  if(packet->type >= PACKET_TYPE_COUNT) {
    return error("Unknown packet type %d", packet->type);
  }

  if(PACKET_HANDLERS[packet->type].client == NULL) {
    return error("Packet type %d has no client handler", packet->type);
  }

  return PACKET_HANDLERS[packet->type].client(packet, client);
}

errorret_t packetServerProcess(
  const packet_t *packet,
  serverclient_t *client
) {
  assertNotNull(packet, "Packet is NULL");
  assertNotNull(client, "Client is NULL");
  assertTrue(
    client->server->type == SERVER_TYPE_NETWORKED,
    "Server is not networked"
  );
  assertIsMainThread("Server client process must be on main thread");

  if(packet->type >= PACKET_TYPE_COUNT) {
    return error("Unknown packet type %d", packet->type);
  }

  if(PACKET_HANDLERS[packet->type].server == NULL) {
    return error("Packet type %d has no server handler", packet->type);
  }

  return PACKET_HANDLERS[packet->type].server(packet, client);
}